#include <iostream.h>
#include <stl.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>

#define PASSER_SEND 0
#define PASSER_REC 1
#define PASSER_FINALIZE 2
#define PASSER_PROBE 3

struct pscux_passer_info_t{
  int req_pipe[2];
  int req2_pipe[2];
  int pipe_data_out[2];
  int pipe2_data_out[2];
  int debug_level;
};

void *PscuxPasserServer( void *inf ){
  pscux_passer_info_t info;
  bool loop , loop2;
  int req;
  int conf;
  
  struct flock fl = { F_WRLCK , SEEK_SET , 0 , 0 , 0 };
  fl.l_pid = getpid();
  
  info = *(pscux_passer_info_t*)inf;
  loop = true;
  conf = 1;

  while(loop){
    /* read a initial req */
    read( info.req_pipe[0] , &req , sizeof(req) );
    /* send confirmatio of req received */
    write( info.pipe_data_out[1] , &conf , sizeof(conf) );
    
    switch(req){
    case PASSER_SEND:
      fl.l_type = F_WRLCK;
      fcntl(info.pipe2_data_out[1], F_SETLKW, &fl);
      /* read a second req( by receiver func) */
      //cout << "read a second req( by receiver func)" << endl;
      read( info.req2_pipe[0] , &req , sizeof(req) );
      //cout << "OK" << endl;
      switch(req){
      case PASSER_REC:
	/* send confirmatio of req received */
	conf = PASSER_REC;
	write( info.pipe2_data_out[1] , &conf , sizeof(conf) );
	break;
      case PASSER_FINALIZE:
	loop = false;
	break;
      default:
	break;
      }
      break;
      
    case PASSER_FINALIZE:
      loop = false;
      break;
    default:
      break;
    }
    fl.l_type = F_UNLCK;
    fcntl(info.pipe2_data_out[1], F_SETLK, &fl);
  }
  return NULL;
}


class PscuxPasser{
 private:
  pscux_passer_info_t info;
  pthread_t thread;
  int req;
  int transfer[2];
  int transfer_conf[2];

 public:
  PscuxPasser(){
    pipe(info.req_pipe);
    pipe(info.req2_pipe);
    pipe(info.pipe_data_out);
    pipe(info.pipe2_data_out);
    pipe(transfer);
    pipe(transfer_conf);
    info.debug_level = 99;
  
    pthread_create( &thread , NULL , PscuxPasserServer , (void*)&info );
  }

  /* send a mess and wait for the receiver confirmation */
  template<class T>  
  void send( T &mess ){
    int conf;
    req = PASSER_SEND;
    write( info.req_pipe[1] , &req , sizeof(req) );
    /* wait confirmation */
    read( info.pipe_data_out[0] , &conf , sizeof(conf) );
    
    /* write the mess */
    //cout << "transferencia iniciada" << endl;
    write( transfer[1] , &mess , sizeof(T) );
    /* wait for confirmation of receive complete */
    //cout << "esperando confirmao de recebimento" << endl;
    read( transfer_conf[0] , &conf , sizeof(conf) );
    //cout << "confirmao de recebimento recebida" << endl;
  }
  
  template<class T>
  void receive( T &mess ){
    int conf;

    req = PASSER_REC;

    /* ask for data */
    write( info.req2_pipe[1] , &req , sizeof(req) );
    /* wait for confirmation */
    read( info.pipe2_data_out[0] , &conf , sizeof(conf) );
    
    /* receiving data ... */
    //cout << "recebendo dados..." << endl;
    read( transfer[0] , &mess , sizeof(T) );
    //cout << "dados recebidos, enviando confirmao" << endl;
    /* send confirmation that the transfer is over */
    write( transfer_conf[1] , &conf , sizeof(conf) );
    //cout << "confirmao enviada" << endl;
  }

  int probe(){
    struct flock fl = { F_UNLCK , SEEK_SET , 0 , 0 , 0 };
    fcntl(info.pipe2_data_out[1], F_GETLK, &fl);
    if(fl.l_type == F_UNLCK){
      return 0;
    }
    return 1;
  }

  void finalize(){
    req = PASSER_FINALIZE;
    write( info.req_pipe[1] , &req , sizeof(req) );
    write( info.req2_pipe[1] , &req , sizeof(req) );
    pthread_join( thread , NULL );
  }
};
