#include <stl.h>
#include <iostream.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/file.h>

#define SH_VAR_SET_VALUE 0
#define SH_VAR_GET_VALUE 1
#define SH_VAR_FINALIZE 2
#define SH_VAR_OK 3

struct pscux_info_t{
  size_t sh_var_size;
  int pipe_req[2];
  int pipe_data_in[2];
  int pipe_data_out[2];
  struct flock fl;
  int debug_level;
};

int PscuxSharedVar_DebugLevel;

void *pscuxShVarServer( void *info_ ){
  cout << "\tservidor SH iniciado" << endl;
  pscux_info_t rec_info;
  void *data;
  int req;
  int conf;
  bool loop;

  loop = true;

  rec_info = *(pscux_info_t*)info_;
  
  data = malloc( rec_info.sh_var_size );
  
  if(rec_info.debug_level > 2){
    cout << "\tmemoria alocada para o valor: " << rec_info.sh_var_size << endl;
    cout << "\tpipe de requerimentos: ( " << rec_info.pipe_req[0] << " , "
	 << rec_info.pipe_req[1] << " )" << endl;
    cout << "\tpipe de entrada de dados: ( " << rec_info.pipe_data_in[0] << " , "
	 << rec_info.pipe_data_in[1] << " )" << endl;
    cout << "\tpipe de saida de dados: ( " << rec_info.pipe_data_out[0] << " , "
	 << rec_info.pipe_data_out[1] << " )" << endl;
  }
  
  while(loop){
    read( rec_info.pipe_req[0] , &req , sizeof(req) );
    // manda confirmaao que recebeu o requerimento (sem alterar req claro)
    write( rec_info.pipe_data_out[1] , &conf , sizeof(conf)); 
    switch (req){
    case SH_VAR_SET_VALUE:
      if(rec_info.debug_level > 2)
	cout << "\trecebi requerimento de set value" << endl;
      read( rec_info.pipe_data_in[0] , data , rec_info.sh_var_size);
      if(rec_info.debug_level > 1)      
	cout << "\tservidor mudou o valor para " << *(int*)data << endl; 
      if(rec_info.debug_level > 2)
	cout << "\tmandando confirmaao que recebeu" << endl;
      req = SH_VAR_OK;
      write( rec_info.pipe_data_out[1] , &conf , sizeof(conf)); 
      break;
    case SH_VAR_GET_VALUE:
      if(rec_info.debug_level > 2)
	cout << "\trecebi requerimento de get value" << endl;
      write( rec_info.pipe_data_out[1] , data , rec_info.sh_var_size);
      // espera confirmao de recebimento
      read( rec_info.pipe_data_in[0] , &conf , sizeof(conf) );
      /*      cout << "confirmado com status " << conf << endl; */
    
      break;
    case SH_VAR_FINALIZE:
      cout << "\trecebi requerimento de finalize" << endl;
      loop = false;
      break;
    default:
      break;
    }
  }
  
  cout << "\tfinalizando Sh Var Server " << pthread_self() << endl;

  //sleep(1);
  pthread_exit(NULL);
  return NULL;
}


template<class T>
class PscuxSharedVar{
 private:
  //T value;
  pscux_info_t info;
  pthread_t thread;
  int req;
  int conf;
  int debug_level;
 public:
  PscuxSharedVar(int dbg_level=0){
    debug_level = dbg_level;
    info.debug_level = dbg_level;
    cout << "debug level = " << debug_level << endl;
    info.sh_var_size = sizeof(T);
    pipe(info.pipe_req);
    pipe(info.pipe_data_in);
    pipe(info.pipe_data_out);
    struct flock fl_temp  = { F_WRLCK , SEEK_SET , 0 , 0 , 0 };
    info.fl = fl_temp;
    pthread_create( &thread , NULL , pscuxShVarServer ,  (void*)&info );
  }
  // muda o valor ( depois rescrever o operador de "=")
  void set_value(T valor){
    req = SH_VAR_SET_VALUE;
    
    write( info.pipe_req[1] , &req , sizeof(req) );
    // espera confirmao
    read( info.pipe_data_out[0] , &conf , sizeof(conf) );
    /*    cout << "set value req confirmado com status " << conf << endl; */

    write( info.pipe_data_in[1] , &valor , sizeof(T) );
    
    if(info.debug_level > 1)
      cout << " vou mandar " << endl;

    // espera confirmao
    read( info.pipe_data_out[0] , &conf , sizeof(conf) );
    /*    cout << "set value finalizao confirmada com status " << conf << endl; */
  }

  T get_value(){
    T temp;
    req = SH_VAR_GET_VALUE;

    write( info.pipe_req[1] , &req , sizeof(req) );
    // espera confirmao
    read( info.pipe_data_out[0] , &conf , sizeof(conf) );
    /*    cout << "get value req confirmado com status " << conf << endl; */
    
    read( info.pipe_data_out[0] , &temp , sizeof(T) );
    // manda confirmao de recebimento do valor
    write( info.pipe_data_in[1] , &conf , sizeof(conf) );
    /*    cout << "get value confirmado com status " << conf << endl; */
    return temp;
  }

  void finalize(){
    //manda sinal de finalizaao
    // espera acabar
    req = SH_VAR_FINALIZE;
    write( info.pipe_req[1] , &req , sizeof(req) );
    pthread_join( thread , NULL );
  }

  void operator= (T x){
    set_value(x);
    //return get_value();
  }
  
  void operator<< (T x){
    set_value(x);
  }

  friend ostream & operator<< ( ostream &s , PscuxSharedVar<T> &ps){
    T temp;
    temp = ps.get_value();
    s << temp;
    return s;
  }
  
  friend istream & operator>> ( istream &s , PscuxSharedVar<T> &ps){
    T temp;
    s >> temp;
    temp = ps.set_value(temp);
    return s;
  }
  
  friend T & operator<< ( T &s , PscuxSharedVar<T> &ps){
    T temp;
    temp = ps.get_value();
    s = temp;
    return s;
  }

  void operator>> (T& x){
    x = get_value();
  }
};
