#include <sys/types.h>
#include <sys/wait.h>
#include <pthread.h> /* threads */
#include <unistd.h> /* pipe() */
#include <stl.h> /* templates */
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include "global.h"
//#include "sh_var.h"
//#include "mutex.h"


/* init procedures */
void pscux_init(){
  /* cria o pipe principal */
  pscux_pipes_ptr = 0;
  
  pscux_servers_on = true;

  pipe( pscux_main_communicator_pipe );
  pipe( pscux_pipe_server_pipe );
  pipe( pscux_pipe_server_pipe_back );
  //  cout << "init pipes created\n";
  
  pthread_create( &pipe_server_thread , NULL , pscux_pipe_server_func , NULL );
  pthread_create( &sh_var_server_thread , NULL , pscux_sh_var_server_func , NULL );
  //cout << "init threads created\n";

}

void pscux_end(){
  pthread_join( pipe_server_thread , NULL );
}

/* main server */
void pscux_server(){
  
}

/* main */
template < class INPUT , class OUTPUT >
class Tarefa
{
public:
  typedef struct d_t{
    INPUT data;
    OUTPUT (*func)(INPUT);
    OUTPUT result;
  }data_t;
  
  data_t data;
  pthread_t thread;
  /* construtor */
  Tarefa( OUTPUT (func_)(INPUT) , INPUT data_ ){
    data.data = data_;
    data.func = func_;
    
  }

  /* Faz uma tarefa no cluster retornando imediatamente         *
   * Recebe um ponteiro pra uma funcao e um ponteiro pros dados */
  int exec(){
    if(pthread_create( &thread , NULL , threadFunc , (void*)&data )){
      return -1;
    }else{
      return 0;
    }
  }
  
  /* espera o thread finalizar e retorna o resultado */
  OUTPUT receive(){
    pthread_join( thread , NULL);
    return data.result;
  }

  static void *threadFunc(void *data){
    data_t *d;
    OUTPUT (*funcao)(INPUT);
    
    d = (data_t *)data ;
    funcao = d->func;
    
    int tubo[2];
    int child_pid;
    OUTPUT temp;
    
    pipe(tubo);
    child_pid = fork();
    if(!child_pid){ /* filho */
      /* executa a funcao */
      temp = funcao(d->data); 
      write(tubo[1],&temp,sizeof(OUTPUT));
      exit(0);
    }else{ /* pai */
      /* le do tubo o que seu filho mandou */
      read(tubo[0],&temp, sizeof(OUTPUT));
      
      /* espera a morte de seu esforcado filho */
      waitpid( child_pid, NULL, 0 );
      
      /* retorna o resultado recebido do filho */
    }
    d->result = temp;
  
  pthread_exit( NULL );
  }
};

// mutex
class pscux_mutex{
 public:
  int associated_pipe[2];
  struct flock fl;

  pscux_mutex(){
    pscux_tpipe temp_pipe;
    temp_pipe = pscux_internal_get_new_tpipe();
    associated_pipe = temp_pipe.pipe;
    //pipe( associated_pipe );
    struct flock fl_temp = {  F_WRLCK, SEEK_SET, 0,       0,     0 };
    fl = fl_temp;
  }
  void lock(){
    cout << "trying to lock: " << associated_pipe[1] << endl;
    fl.l_pid = getpid();
    fcntl(associated_pipe[1], F_SETLKW, &fl);
  }
  void unlock(){
    //cout << "trying to unlock: " << associated_pipe[1] << endl;
    fl.l_pid = getpid();
    fl.l_type = F_UNLCK;
    fcntl(associated_pipe[1], F_SETLK, &fl);
  }
};


// globals
#define PSCUX_MAX_PROCS 100


template< typename T>
size_t pscux_sizeof( T& x ){
  return sizeof(x);
}


// sh var
#define PSCUX_MAX_SH_VARS 10

typedef struct pscux_sh_var_server_req{
  int type;
  int pos; /* qual pos alterar */
}pscux_sh_var_server_req;

typedef struct pscux_generic_t{
  void *value;
  size_t size;
}pscux_generic_t;

int pscux_sh_var_server_pipe[2];
int pscux_sh_var_server_pipe_back[2];

int global_sh_var_pos=0;

class Psh_vars{
 public:
  pscux_generic_t sh_vars[PSCUX_MAX_SH_VARS];
  int sh_vars_ptr;

  Psh_vars(){
    sh_vars_ptr = 0;
  }

  template< class T >
    void set_value( int pos , T& new_value ){
    /* pode dar warning em diferenca de tamanho e erro em out_of_range */
    free( sh_vars[pos].value );

    sh_vars[ pos ].value = malloc( pscux_sizeof(new_value) );
    sh_vars[ pos ].size = pscux_sizeof(new_value) ;

    memcpy( sh_vars[pos].value , (void*)&new_value , pscux_sizeof(new_value) );
  }
  
  void gset_value( int pos , pscux_generic_t g ){
    free( sh_vars[pos].value );
     
    //cout << "\tgset_value:" << endl
    //	 << "\t\tsize: " << g.size << endl
    //	 << "\t\tvalue: " << *(int*)g.value << endl;
    
    sh_vars[ pos ].value = malloc( g.size );
    sh_vars[ pos ].size = g.size ;

    memcpy( sh_vars[pos].value , g.value , g.size );
    //    cout << "\t\tvalue after: " << *(int*)sh_vars[pos].value << endl;
  }
  
  template< class T >
    void get_value( int pos , T* dest ){
    /* same warnings maybe */
    memcpy( (void*)dest , sh_vars[pos].value , sh_vars[pos].size );
  }
  
  void gget_value( int pos , pscux_generic_t& dest ){
    memcpy( dest.value , sh_vars[pos].value , sh_vars[pos].size );
    memcpy( (void*)&dest.size ,  (void*)&sh_vars[pos].size , pscux_sizeof(sh_vars[pos].size) );
  }

  template< class T >
    void add_new( T& value ){
    sh_vars[ sh_vars_ptr ].value = malloc( sizeof(T) );
    sh_vars[ sh_vars_ptr ].size = sizeof(T) ;

    memcpy( sh_vars[sh_vars_ptr].value , (void*)&value , pscux_sizeof(value) );
    sh_vars_ptr++;
  }

  void gadd_new( pscux_generic_t g ){
    sh_vars[ sh_vars_ptr ].value = malloc( g.size );
    sh_vars[ sh_vars_ptr ].size = g.size ;

    memcpy( sh_vars[sh_vars_ptr].value , (void*)&g.value , g.size );
    sh_vars_ptr++;
  }
};

/* servidor de transaes */
void *pscux_sh_var_server_func( void *data ){
  Psh_vars shared_vars;  
  pscux_sh_var_server_req req;
  pscux_generic_t generic_temp;
  int pos;
  size_t size;

  //  int temp;
  
  pipe( pscux_sh_var_server_pipe );
  pipe( pscux_sh_var_server_pipe_back );

  while( pscux_servers_on ){
    /* se no der certo com o servers_on fazer com um valor do req */
    read( pscux_sh_var_server_pipe[0] , &req , pscux_sizeof(req) );
    pos = req.pos;

    //cout << "sh_var_server req receiveid:" << endl
    //	 << "\ttype: " << req.type << endl
    //	 << "\tpos: " << req.pos
    //	 << endl;

    if(req.type == 1){
      /* add_new */
      //cout << "req type 1" << endl;
      read( pscux_sh_var_server_pipe[0] , &size , pscux_sizeof(size) );
      generic_temp.value = malloc(size);
      generic_temp.size = size;
      read( pscux_sh_var_server_pipe[0] , generic_temp.value , pscux_sizeof(size) );
      shared_vars.gadd_new( generic_temp );
    }else if(req.type == 2){
      /* set_value */
      //cout << "req type 2" << endl;
      read( pscux_sh_var_server_pipe[0] , &size , pscux_sizeof(size) );
      generic_temp.value = malloc(size);
      generic_temp.size = size;
      read( pscux_sh_var_server_pipe[0] , generic_temp.value , pscux_sizeof(size) );
      //temp = *(int*)generic_temp.value;
      // cout << "generic_temp:" << endl
      //	   << "\tvalue: " << *(int*)generic_temp.value << endl
      //	   << "\tsize: " << generic_temp.size << endl
      //	   << "\ttemp: " << temp << endl;
      shared_vars.gset_value( pos , generic_temp );
    }else if(req.type == 3){
      /* get_value */
      //cout << "req type 3" << endl;
      read( pscux_sh_var_server_pipe[0] , &size , pscux_sizeof(size) );
      generic_temp.value = malloc(size);
      generic_temp.size = size;
      shared_vars.gget_value( pos , generic_temp );
      write( pscux_sh_var_server_pipe_back[1] , generic_temp.value , size );
    }
    free(generic_temp.value);
  }     
  pthread_exit( NULL );
}

/* definio de uma variavel compartilhada */
template< class T >
class pscux_shared_var{
 public:
  int position;
  pscux_mutex mutex;
  pscux_sh_var_server_req req;
  size_t size;  
  
  pscux_shared_var(){
    /* call add new trougth reads and writes */
    position = global_sh_var_pos++;
  }

  void init(){
    //    sleep(1);
    T temp;
    size = sizeof(temp);
    // cout << "pos: " << position << endl;
    req.type = 1; /* add new */
    req.pos = position;
    //    cout << "set_value_send_req:" << endl
    // << "\ttype: " << req.type << endl
    //	 << "\tpos: " << req.pos << endl
    //	 << "\tsize: " << sizeof(req) << endl;
    mutex.lock();
    write( pscux_sh_var_server_pipe[1] , (void*)&req , sizeof(req) );
    write( pscux_sh_var_server_pipe[1] , &size , sizeof(size_t) );
    write( pscux_sh_var_server_pipe[1] , &temp , sizeof(temp) );
    mutex.unlock();
  } 
  void set_value( T& new_value ){
    //   cout << "set_value to: " << new_value << endl;
    size = sizeof(new_value);
    // cout << "set_value size: " << size << endl;
    req.type = 2;
    req.pos = position;
    // cout << "set_value_send_req:" << endl
    //	 << "\ttype: " << req.type << endl
    //	 << "\tpos: " << req.pos << endl
    //	 << "\tsize: " << sizeof(req) << endl;
    mutex.lock();
    write( pscux_sh_var_server_pipe[1] , (void*)&req , sizeof(req) );
    write( pscux_sh_var_server_pipe[1] , (void*)&size , sizeof(size_t) );
    //cout << "WRITE: " << new_value << endl;
    write( pscux_sh_var_server_pipe[1] , &new_value , pscux_sizeof(new_value) );
    mutex.unlock();
  }
  
  T get_value(){
    T temp;
    req.type = 3;
    req.pos = position;
    size =  sizeof(T);
    //cout << "set_value_send_req:" << endl
    //	 << "\ttype: " << req.type << endl
    //	 << "\tpos: " << req.pos << endl
    //	 << "\tsize: " << sizeof(req) << endl;
    mutex.lock();
    write( pscux_sh_var_server_pipe[1] , (void*)&req , sizeof(req) );
    write( pscux_sh_var_server_pipe[1] , (void*)&size , sizeof(size_t) );
    read( pscux_sh_var_server_pipe_back[0] , &temp , sizeof(T) );
    //cout << "get_value to: " << temp << endl;
    return temp;
    mutex.unlock();
  }
};

// pipe server



typedef struct pscux_pipe_server_req{
  int type;
}pscux_pipe_server_req;

int pscux_pipes[PSCUX_MAX_PROCS][2];

pscux_tpipe pscux_internal_get_new_tpipe(){
  /* lock 1 */
  pscux_tpipe temp_pipe;
  pipe( pscux_pipes[ pscux_pipes_ptr ] );
  temp_pipe.pipe = pscux_pipes[ pscux_pipes_ptr ];
  //cout << "internal_get_new_pipe: generated pipe[0]-> "
  //    << temp_pipe.pipe[0] << endl
  //    << "internal_get_new_pipe: generated pipe[1]-> "
  //    << temp_pipe.pipe[1] << endl;
  pscux_pipes_ptr++;
  /* unlock 1 */
  return temp_pipe;
}


/* servidor de pipes */
void *pscux_pipe_server_func( void *data ){
  pscux_tpipe temp_pipe;
  pscux_pipe_server_req req;
  while( pscux_servers_on ){
    read( pscux_pipe_server_pipe[0] , &req , pscux_sizeof(req) );
    // cout << "pipe_server read req-> " << req.type << endl;
    temp_pipe = pscux_internal_get_new_tpipe();
    write( pscux_pipe_server_pipe_back[1] , &temp_pipe , pscux_sizeof(temp_pipe) );
  }
  // cout << "pipe_server" << endl;
  pthread_exit( NULL );
}

/* end-user function */
void pscux_get_new_pipe( int* p ){
  pscux_tpipe temp_pipe;
  pscux_pipe_server_req req;

  req.type = 1; /* get_new */
  write( pscux_pipe_server_pipe[1] , &req , pscux_sizeof(req) );
  read( pscux_pipe_server_pipe_back[0] , &temp_pipe , pscux_sizeof(temp_pipe) );
  memcpy( p , temp_pipe.pipe , pscux_sizeof(temp_pipe) );
}

